/*
			Copyright (c) 1997 by
			Advanced Visual Systems Inc.
			All Rights Reserved

	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.

	This file is under Perforce control
	$Id: //depot/express/fcs992/modules/rd_ucd2.c#1 $
*/

/********************************************************************

	file name:	rd_ucd2.c
	module name:	Read_UCD
	update:		1.27.98
	authour:	Y. Kikkawa (Original: Y. Nakai)
			(KGT Inc.)

 ********************************************************************/

#include <stdio.h>

#include <avs/util.h>
#include <avs/err.h>
#include <avs/om.h>
#include <avs/math.h>
#include <avs/dtype.h>
#include <avs/fld.h>
#include <avs/f_utils.h>

#define		BUFFER_SIZE	1024 /* max number of characters in a line */
			   	     /* in the data file */
#define		ERR_RETURN(A)	ERRerror("Read_UCD",1,ERR_ORIG, A); return(0);
#define		BLOCK_SIZE	512

static int skip_geom_data(FILE *fp, int *num_nodes, int *num_cells);

struct cell {
  int		mat_id;
  int		type_id;
  int		con_num;
  int		cell_num;
  int		size;
  int		*connect;
  int		*cell_no;
  struct cell	*next;
};


/**********************************************************************
 *
 *	function name:	get_token
 *
 *	description:	extract the next string in tline[] separeted by
 *		 	blank and  convert it in the specified form
 *
 *	return: ntc	size of extracted string 
 * 
 **********************************************************************/
static int
get_token (
char	**tline,	/* string (input) */
char	*data,		/* extracted and converted string (output) */
char	*control	/* format (e.g. %d, %f) (input) */
)
{
    char  *line, token[80], c;
    int   ntc;

    line = *tline;

    c = 0;
    for (ntc = 0; c != '\n'; line++) {
        if (((c  = *line) != ' ') && (c != '\t') && (c != '\f') && (c != '\n')) {
            token[ntc++] = c;
        } else if (ntc != 0) {
            token[ntc] = '\0';
            sscanf (token, control, data);
            *tline = line;
            break;
        }
    }
    return (ntc);
}


/**********************************************************************
 *
 *	function name:  get_label
 *
 *      description: 	get label and unit of data component
 *			from string separeted by delimiter
 *
 *	return:		1: success 
 *			0: failure
 *
 **********************************************************************/
static int
get_label (
char	*string,		/* string (input) */
char	delimiter,		/* delimiter (input) */
int	number,			/* component number (output) */
char	*label			/* label or unit (output) */
)
{
  int   i, j, k, len, count;
  char  current;

  if (string == NULL)
    return 0;

  count = 1; k = 0;
  len = strlen (string);
  for (i = 0; i <= number; i++) {
    current = string[k++];
    j = 0;
    while (current != delimiter) {
      /* build the label character by character */
      label[j++] = current;
      current = string[k++];

      /* the last character was found */
      if (k > len) {
        /* the nth label was not found, where n = number */
        if (i < number) {
          return 0;
        }
        current = delimiter;
      }

    }
    label[j] = '\0';
  }
  return 1;
}

/**********************************************************************
 *
 *	function name:  ucd_qsi
 *
 *      description:	sort data
 *
 *	return:		1: success 
 *			0: failure
 *
 **********************************************************************/
static int
ucd_qsi(
int	*values,
int	*indexes,
int	left,
int	right
)
{
  register int i, j;
  int x, y;
  int z;

  x = values[(left + right) / 2];
  i = left;
  j = right;

  do {
    while (values[i] < x && i < right) i++;
    while (x < values[j] && j > left) j--;

    if (i <= j) {
      y = values[i];
      values[i] = values[j];
      values[j] = y;
      z = indexes[i];
      indexes[i] = indexes[j];
      indexes[j] = z;
      i++;
      j--;
    }
  } while (i <= j);

  if (left < j) ucd_qsi (values, indexes, left, j);
  if (i < right) ucd_qsi (values, indexes, i, right);
  return 1;
}

/**********************************************************************
 *
 *	function name:  ucd_bsearch
 *
 *      description: 	search
 *			(calculate the index of v in list[])
 *
 *	return:		calculated index
 *
 **********************************************************************/
static int
ucd_bsearchi (
int	v,			/* value that will be serched (input) */
int	*list,			/* dimension of values (input) */
int	n			/* size of list (input) */
)
{
  int found, top, bot, indexer;

  n = n - 1;

  if (v <= list[0])
    return (0);

  if (v >= list[n])
    return (n);

  found = 0;
  top = 0;
  bot = n;

  while (!found) {
    indexer = (top + bot) / 2;

    if (v > list[indexer])
      top = indexer;
    else if (v < list[indexer])
      bot = indexer;
    else
      return (indexer);

    if ((bot - top) < 2)
      if (fabs(list[bot] - v + 0.0) < fabs(v - list[top] + 0.0))
        return (bot);
      else
        return (top);
  }
  return 0;
}

/**********************************************************************
 *
 *	function name:  ucd_qsorti
 *
 *      description:	sort data
 *
 *	return:		1: success
 *			0: failure
 *
 **********************************************************************/
static int
ucd_qsorti(
int	*values,
int	*indexes,
int	num_entries
)
{
  int *x;
  int *ix, mid;
  register int i, j, k;

  mid = (num_entries - 1) / 2;
  if (mid == 0) mid = 1;

  x = (int *)malloc(mid *sizeof(int));
  ix = (int *)malloc(mid * sizeof(int));

  /* dir no_recurrence */
  for (i = 0; i < mid; i++) {
    x[i] = values[i];
    ix[i] = indexes[i];
  }

  /* dir fork */
  ucd_qsi (x, ix, 0, mid-1);
  ucd_qsi (values, indexes, mid, num_entries-1);
  /* dir join */

  i = 0; j = mid;

  while ((i < mid) && (j < num_entries)) {
    if (values[j] < x[i]) {
      values[i + j - mid] = values[j];
      indexes[i + j - mid] = indexes[j];
      j++;
    }
    else {
      values[i + j - mid] = x[i];
      indexes[i + j - mid] = ix[i];
      i++;
    }
  }

  if (i < mid) {
    /* dir no_recurrence */
    for (k = i; k < mid; k++) {
      values[k + j - mid] = x[k];
      indexes[k + j - mid] = ix[k];
    }
  }

  free(x);
  free(ix);
  return 1;
}

/**********************************************************************
 *
 *	function name:  delete_cell_inf
 *
 *      description:	free cell information data
 *
 *	return:		1: success 
 *
 **********************************************************************/
static int
delete_cell_inf(
struct cell	**Top
)
{
  struct cell	*p, *del;

  del = (*Top);
  while( del != NULL ) {
    if( del->cell_no != NULL ) {
      free( del->cell_no );
    }
    if( del->connect != NULL ) {
      free( del->connect );
    }
    p = del->next;
    free( del );
    del = p;
  }
  (*Top) = NULL;
  return 1;
}

/**********************************************************************
 *
 *	function name:  reset_out_data
 *
 *      description:	initialize ouput ucd data
 *
 *	return:		1: success
 *			0: failure 
 *
 **********************************************************************/
static int
reset_out_data(
OMobj_id	out,	/* id of ucd (fld) data object (input) */
int		time_flag	/* 0: single field,  
			  	   1: time-dependent field */
)
{
  int      stat, i, j, k, ncell_sets, ncomps, type, size;
  char     *data;
  int      *conn_array;
  float    *coord;
  OMobj_id cell_set_id, cell_id;

  if (time_flag == 1)
  {
    int      nsteps;
    OMobj_id time_id, time_coord_id;
    double   *time;

    if ((FLDget_time_nsteps(out, &nsteps) != 1) || nsteps <= 0)
      return 0;

    for (i = 0; i < nsteps; i++)
    {
      if ((FLDget_cell_set(out, 0, &cell_set_id) == 1) &&
          (FLDget_cell_data_ncomp(cell_set_id, &ncomps) == 1) &&
          (FLDget_ncell_sets(out, &ncell_sets) == 1))
      {
        for (j = 0; j < ncell_sets; j++)
        {
          if (FLDget_cell_set(out, j, &cell_set_id) == 1)
            stat = FLDset_cell_data_ncomp(cell_set_id, 0);
        }
      }

      if (FLDget_time_node_data(out, i, &time_id) == 1)
      {
        stat = FLDset_nnodes(time_id, 0);
        stat = FLDset_nspace(time_id, 0);
        stat = FLDset_node_data_ncomp(time_id, 0);
      }

      if (FLDget_time_coordinates(out, i, &time_coord_id) != 1)
        stat = FLDset_coord(time_coord_id, NULL, 0, OM_SET_ARRAY_STATIC);
    }
    stat = FLDset_ncell_sets(out, 0);
    stat = FLDset_time_nsteps(out, 0);
  }
  else
  {
    if ((FLDget_cell_set(out, 0, &cell_set_id) == 1) &&
        (FLDget_cell_data_ncomp(cell_set_id, &ncomps) == 1) &&
        (FLDget_ncell_sets(out, &ncell_sets) == 1))
    {
      for (j = 0; j < ncell_sets; j++)
      {
        if (FLDget_cell_set(out, j, &cell_set_id) == 1)
          stat = FLDset_cell_data_ncomp(cell_set_id, 0);
      }
    }

    stat = FLDset_nnodes(out, 0);
    stat = FLDset_nspace(out, 0);
    stat = FLDset_coord(out, NULL, 0, OM_SET_ARRAY_STATIC);
    stat = FLDset_node_data_ncomp(out, 0);
    stat = FLDset_ncell_sets(out, 0);
  }

  return 1;
}


/**********************************************************************
 *
 *	function name:  read_header
 *
 *      description:    skip #lines and read num_steps and cycle_type
 *
 *	return:		1: success 
 *			0: failure 
 *			-1: binary, must be single-step
 *
 **********************************************************************/
/* from modules/rd_ucd.c: get it in shared header file */
#define UCD_ASCII_FILE 3

static int
read_header(
char    *filename,
FILE	*fp,
int	*num_steps,
char	*cycle_type
)
{
    char  line[BUFFER_SIZE];

    /* check to be sure it's ASCII */
    if (FUNCcheck_ucd_file(filename) == UCD_ASCII_FILE) {

      /* skip comment lines which starts from # */
      while (fgets(line, sizeof(line), fp) != NULL) {
        if (line[0] != '#') {
          break;
        }
      }

      if (sscanf(line, "%d", num_steps) != 1) {
        return 0;
      }

      if (fgets(line, sizeof(line), fp) == NULL) {
        return 0;
      }

      if (sscanf( line, "%s", cycle_type) != 1) {
        return 0;
      }
      return 1 ;
    }
    else
      return -1 ;
}


/**********************************************************************
 *
 *	function name:  skip_header
 *
 *      description:    skip #lines and num_steps and cycle_type
 *
 *	return:		1: success 
 *
 **********************************************************************/
static int
skip_header(
FILE	*fp
)
{
    char  line[BUFFER_SIZE];


    /* skip comment lines which starts from # */
    while (fgets(line, sizeof(line), fp) != NULL) {
        if (line[0] != '#') {
            break;
        }
    }


    fgets (line, sizeof(line), fp);

    return 1 ;
}


/**********************************************************************
 *
 *	function name:  read_step_comment
 *
 *      description:    check step and read comment string
 *
 *	return:		1: success 
 *			0: failure 
 *
 **********************************************************************/
static int
read_step_comment(
FILE	*fp,
int	step_no,	/* step number */
char	*comment
)
{
   char	line[BUFFER_SIZE];
   char	step[20];
   char check[20];
   int  len, i, j;


   if (fgets(line, sizeof(line), fp) == NULL) {
       return 0;
   }
   if (sscanf(line, "%s", step) != 1) {
       return 0;
   }

   sprintf(check, "step%d", step_no);

   if (strcmp(step, check) != 0) {
       return 0;
   }

   len = strlen(step);

   sscanf(line, "%s %s", step, comment);
   return 1;
}


/**********************************************************************
 *
 *	function name:	get_step_file_pos 
 *
 *	description:  	get file pointer of beginning of each 
 *                      step
 *
 *	return:		1: success 
 *
 **********************************************************************/
static int
get_step_file_pos(
FILE	*fp,
int	num_steps,
char	*cycle_type,
long	**file_pos	/* position of beginning of each step */
)
{
    char	line[BUFFER_SIZE], err_mes[128];
    int	set_num, num, num_node_data, num_cell_data, num_comp, dummy;
    int	i, j, offset, len;
    int	num_nodes, num_cells;


    /* allocation of file pointer array */
    *file_pos = (long *)malloc(sizeof(long)*(num_steps+1));
    if (*file_pos == NULL) {
        ERR_RETURN("Can't alloc memory.(file_pos)");
    }

    rewind(fp);

    /* skip header */
    skip_header(fp);

    /* get file pointers of beginning of all steps */
    for (set_num = 0; set_num < num_steps; set_num++) {

	/* set current file pointer */
	(*file_pos)[set_num] = ftell(fp);

	/* skip step ID */
	if (fgets(line, sizeof(line), fp) == NULL) {
	    sprintf(err_mes, "Error in step %d: Can't find step_id line.",
		 set_num);
	    ERR_RETURN(err_mes);
	}

	/* skip geometry data */
	if (set_num == 0 || strcmp(cycle_type, "geom") == 0 ||
			strcmp(cycle_type, "data_geom") == 0) {
	    if (skip_geom_data(fp, &num_nodes, &num_cells) != 1) {
		sprintf(err_mes,
			"Error in step %d: Can't skip geometry",
			set_num);
		ERR_RETURN(err_mes);
	    }
	}

        /* set position of beginning of node and cell data of first step  */
        /* to file_pos array (index: num_steps) */
	if (set_num == 0) {
	    (*file_pos)[num_steps] = ftell(fp);
	}

        /* skip node and cell data */
	if (set_num == 0 || strcmp(cycle_type, "data") == 0 ||
			strcmp(cycle_type, "data_geom") == 0) {

	    /* get num_node_data and num_cell_data */
	    if (fgets(line, sizeof(line), fp) == NULL) {
		sprintf(err_mes,
			"Error in step %d: Can't find num_node_data and num_cell_data line.",
			set_num);
		ERR_RETURN(err_mes);
	    }
	    num = sscanf(line, "%d %d", &num_node_data, &num_cell_data);
	    if (num != 2) {
		sprintf(err_mes,
			"Error in step %d: Can't read num_node_data and num_cell_data.",
			set_num);
		ERR_RETURN(err_mes);
	    }

	    /* skip node data */
	    if (num_node_data > 0) {
	        /* get num_comp */
	        if (fgets(line, sizeof(line), fp) == NULL) {
	  	    sprintf(err_mes,
			"Error in step %d: Can't find num_comp line.",
			set_num);
		    ERR_RETURN(err_mes);
	        }
	        sscanf(line, "%d", &num_comp);

	        /* skip node data and component label data */
	        offset = num_nodes + num_comp;
	        for (i = 0; i < offset; i++) {
	    	    fgets(line, sizeof(line), fp);
	        }
	    }

	    /* skip cell data */
	    if (num_cell_data > 0) {
	        /* get num_comp */
	        if (fgets(line, sizeof( line ), fp)  == NULL) {
	    	    sprintf(err_mes,
			"Error in step %d: Can't find num_comp line.",
			set_num);
		    ERR_RETURN(err_mes);
	        }
	        sscanf(line, "%d", &num_comp);

	        /* skip cell data and component label data */
	        offset = num_cells + num_comp;
	        for (i = 0; i < offset; i++) {
		    fgets(line, sizeof(line), fp);
		}
	    }
	}

    }
    return 1;
}


/**********************************************************************
 *
 *	function name:	skip_geom_data 
 *
 *	description: 	skip geometry data
 *
 *	return:		1: success 
 *			0: 
 *
 **********************************************************************/
static int
skip_geom_data (
FILE	*fp,
int	*num_nodes,
int	*num_cells
)
{
  char	line[BUFFER_SIZE], err_mes[128];
  int	i;

    /* get num_nodes and num_cells */
    fgets( line, sizeof( line ), fp );
    if (sscanf(line, "%d %d", num_nodes, num_cells) != 2) {
        sprintf( err_mes,
		"Can't get num_nodes or num_cells of the first step.");
        ERR_RETURN( err_mes );
    }
    /* skip geometry data */
    for (i = 0; i < *num_nodes + *num_cells; i++) {
        fgets( line, sizeof( line ), fp );
    }
    return 1;
}


/**********************************************************************
 *
 *	function name:  read_geom_data
 *
 *      description:	read geometry data
 *
 *	return:		1: success 
 *			0: failure
 *
 **********************************************************************/

#define	NUM_CELL_TYPE	15

static int
read_geom_data(
FILE		*fp,		/* file pointer */
int		*num_nodes,	/* number of nodes */
int		*num_cells,	/* number of cells */
OMobj_id	out,		/* id of ucd output object */
struct cell	**Top,		/* cell information */
int		time_step,	/* time step: if 0, output is single field */
int		num_steps 	/* number of steps of time-dependent field */
)
{
    static int	con_nodes[] = { 1, 2, 3, 4,  4,  5,  6,  8,
				   3, 6, 8, 10, 13, 15, 20 };
    static char	*type_list[] = { "pt", "line", "tri", "quad",
				 "tet", "pyr", "prism", "hex",
				       "line2", "tri2", "quad2",
				 "tet2", "pyr2", "prism2", "hex2" };
    static char	*set_type_list[] = { "Point", "Line", "Tri", "Quad",
				     "Tet", "Pyr", "Prism", "Hex",
				             "Line2", "Tri2", "Quad2",
				     "Tet2", "Pyr2", "Prism2", "Hex2" };
    char	line[BUFFER_SIZE], *tl, err_mes[128];
    char	type[BUFFER_SIZE], coord_units[10];
    int	i, j, k, cnt, type_len, n, num;
    int	nsets, cell_no, mat, alloc_cnt, m, node, cs;
    int	*node_ids, *node_ind, ncell_sets;
    float	*coord;
    OMobj_id	cell_set;
    struct cell	*p, *last;
    float	*props;
    int		size;
    OMobj_id	coord_id;
    int		first_num_nodes, first_ncell_set;


    /* get num_nodes and num_cells */
    if (fgets (line, sizeof(line), fp) == NULL) {
	ERR_RETURN("Can't get num_nodes and num_cells");
    }
    if (sscanf(line, "%d %d", num_nodes, num_cells) != 2) {
	ERR_RETURN("Can't get num_nodes or num_cells");
    }

    if (time_step < 2) {
	/* set num_nodes */
	FLDset_nnodes (out, *num_nodes);

	/* set nspace */
	FLDset_nspace (out, 3);
    }

    if (time_step >= 2) {
        if (FLDget_nnodes (out, &first_num_nodes) != 1) {
	    ERR_RETURN("Can't get first_num_nodes.")
        }
	if (*num_nodes != first_num_nodes) {
	    ERR_RETURN("Can't read this data under Store All Steps mode because numbers of nodes don't match between steps.")
        }
    }

    /*******************************************/
    /* get and set node coodinates and node id */
    /*******************************************/

    if (time_step >= 1) {
        if (FLDget_time_coordinates (out, time_step-1, &coord_id) != 1) {
	    ERR_RETURN("Can't get coord_id of output time-dependent field.");
        }
    } else {
	coord_id = out;
    }


    /* set coord unit */
    strcpy(coord_units, "");
    FLDset_coord_units(coord_id, coord_units);

    /* allocation */
    node_ids = (int *)malloc(sizeof(int) * *num_nodes);
    if (node_ids == NULL) {
	ERR_RETURN("Can't alloc memory.(node_ids)");
    }

    node_ind = (int *)malloc(sizeof(int) * *num_nodes);
    if (node_ind == NULL) {
        free( node_ids );
        ERR_RETURN("Can't alloc memory.(node_ind)");
    }

    coord = (float *)ARRalloc(NULL, DTYPE_FLOAT,
				*num_nodes*3, NULL);
    if (coord == NULL) {
        free(node_ids);
        free(node_ind);
        ERR_RETURN("Can't alloc memory.(coord)");
    }

    /* get node id and coordinates */
    for (i = cnt = 0; i < *num_nodes; i++, cnt+=3) {
	fgets(line, sizeof( line ), fp);
	num = sscanf(line, "%d %f %f %f", &node_ids[i],
		  &coord[cnt], &coord[cnt+1], &coord[cnt+2]);
	if (num != 4) {
	    free(node_ids);
            free(node_ind);
            ERR_RETURN("Can't read node coordinate.");
        }
	node_ind[i] = i;
    }

    /* sort node ids */
    ucd_qsorti(node_ids, node_ind, *num_nodes);

    /* set coordinates */
    if (FLDset_coord (coord_id, coord, *num_nodes*3, OM_SET_ARRAY_FREE) != 1) {
        ERR_RETURN("Can't set coordinate.");
    }

    /*****************************/
    /* get and set cell topology */
    /*****************************/
    if (time_step <= 1) {
        delete_cell_inf(Top);

	ncell_sets = 0;
	for (i = 0; i < *num_cells; i++) {
	    /* get cell_id, material_id and cell_type */
	    if (fgets( line, sizeof( line ), fp ) == NULL) {
		free( node_ids );
		free( node_ind );
		ERR_RETURN("Can't read cell connectivity.");
	    }
	    tl = line;
	    n = 0;

	    get_token (&tl, (void *) &cell_no, "%d");
	    get_token (&tl, (void *) &mat, "%d");
	    type_len = get_token (&tl, (void *)type, "%s");
	    type[type_len] = '\0';

	    /* check cell type */
	    for (k = 0; k < NUM_CELL_TYPE; k++) {
		if (!strncmp( type, type_list[k], type_len)) {
		    break;
		}
	    }
	    if (k == NUM_CELL_TYPE) {
		sprintf(err_mes, "Can't support cell_type[%s].", type);
		free(node_ids);
		free(node_ind);
		ERR_RETURN(err_mes);
	    }

	    for (p = *Top; p != NULL; p = p->next) {
		if (p->mat_id != mat) {
		    continue;
		}
		if (p->type_id != k) {
		    continue;
		}
		break;
	    }
	    if (p == NULL) {
		p = (struct cell *)malloc(sizeof(struct cell));
		if (p == NULL) {
		    free (node_ids);
		    free( node_ind );
		    ERR_RETURN("Can't alloc memory.(struct cell)");
		}
		memset(p, 0, sizeof(struct cell));

		if (*Top == NULL) {
		    *Top = last = p;
		} else {
		    last->next = p;
		    last = p;
		}

		p->mat_id = mat;
		p->type_id = k;
		p->con_num = con_nodes[k];

		p->cell_num = 1;

		ncell_sets++;
	    } else {
		p->cell_num++;
	    }

	    if (!p->cell_no) {
		p->size = BLOCK_SIZE;
		p->cell_no = (int *)malloc(p->size*sizeof(int));
		p->connect = (int *)malloc(p->con_num*p->size*sizeof(int));
	    } else if (p->cell_num > p->size) {
		p->size += BLOCK_SIZE;
		p->cell_no = (int *)realloc(p->cell_no, p->size*sizeof(int));
		p->connect = (int *)realloc(p->connect,
					    p->con_num*p->size*sizeof(int));
	    }

	    if (p->cell_no == NULL || p->connect == NULL) {
		if (p->cell_no) free(p->cell_no);
		if (p->connect) free(p->connect);
		free( node_ids );
		free( node_ind );
		ERR_RETURN("Can't alloc memory.(p->cell_no or p->connect)");
	    }

	    p->cell_no[p->cell_num-1] = i;


	    if (p->connect == NULL) {
		free(node_ids);
		free(node_ind);
		ERR_RETURN("Can't alloc memory.(p->connect)");
	    }

	    for (m = 0; m < p->con_num; m++) {
		if (get_token(&tl, (void *)&node, "%d") < 1) {
		    free(node_ids);
		    free(node_ind);
		    ERR_RETURN("Can't read cell connectivity");
		}

		k = ucd_bsearchi(node, node_ids, *num_nodes);

		alloc_cnt = p->cell_num * p->con_num;
		p->connect[alloc_cnt-p->con_num+m] = node_ind[k];
	    }
	}

	free(node_ids);
	free(node_ind);

	cell_set = OMfind_subobj(out, OMstr_to_name("cell_set"), OM_OBJ_RW);
	OMreset_obj(cell_set, 0 );

	FLDset_node_data_ncomp(out, 0);

	FLDset_ncell_sets(out, ncell_sets);

	for (p = *Top, cs=0; p != NULL; p = p->next) {
	    FLDget_cell_set(out, cs, &cell_set);

	    FLDset_cell_set(cell_set, set_type_list[p->type_id]);

	    FLDset_ncells(cell_set, p->cell_num );

	    FLDset_node_connect(cell_set, p->connect, p->cell_num*p->con_num,
			    OM_SET_ARRAY_COPY);

	    /* set material ID */
	    if (FLDset_cell_nprops (cell_set, 1) != 1) {
		ERR_RETURN("Can't set nprops");
	    }
	    if (FLDget_cell_props(cell_set, (float **)&props, &size,
		    OM_GET_ARRAY_WR) != 1) {
		ERR_RETURN("Can't get cell props");
	    }
	    props[0] = p->mat_id;
	    ARRfree(props);

	    cs++;

	    /* set nsteps */
            if (time_step == 1) {  /* in case of time-dependent data */
	        if (FLDset_time_nsteps(cell_set, num_steps) != 1) {
	            ERR_RETURN("Can't set nsteps of cell_set of time-dependent field.");
                }
            }
	}
    }

    if (time_step >= 2) {
        /* skip cell connectivity */
	for (i = 0; i < *num_cells; i++) {
	    if (fgets( line, sizeof( line ), fp ) == NULL) {
		ERR_RETURN("Can't read cell connectivity.");
            }
        }
    }
    return 1;
}


/**********************************************************************
 *
 *	function name:	read_data
 *
 *	description: 	read node / cell data
 *
 *	return:		1: success 
 *			0: failure 
 *
 **********************************************************************/
static int
read_data(
FILE 	*fp,			/* file pointer (input) */
int 	num_data,		/* number of node or cell (input) */
int 	sum_veclen,		/* sum of veclen of each component */
float 	*data,			/* data (output) */
char 	*data_labels,		/* data label (ouput) */
char 	*units,			/* data unit (ouput) */
int	*num_comp,		/* number of component (ouput) */
int 	*comp_list		/* component list (output) */
)
{
  char	line[80];
  int 	offset, num_vals, node, i, j, n, m, id, num;
  float value;
  int 	sum;

  num = fscanf (fp, "%d", num_comp);
  if( num != 1 ) {
    ERR_RETURN("Can't read step component number");
  }
  if (*num_comp > sum_veclen) {
    ERR_RETURN("Number of components is greater than number of data.");
  }


  sum = 0;
  for (i = 0; i < *num_comp; i++) {
    num = fscanf (fp, "%d", &comp_list[i]);
    if( num != 1 ) {
      ERR_RETURN("Can't read step comp_list.");
    }
    sum += comp_list[i];
  }
  if (sum != sum_veclen) {
      ERR_RETURN("Sum of veclen is not coincident with number ob data.");
  }

  fgets (line, BUFFER_SIZE, fp);

  for (i = 0, n = 0, m = 0; i < *num_comp; i++) {

    if (fgets (line, BUFFER_SIZE, fp) == NULL) {
      ERR_RETURN("Can't read comp_label and unit_label");
    }

    for (j = 0; line[j] == ' '; j++);

    for (; line[j] != ','; j++) {
      data_labels[n++] = line[j];
      if (n >= BUFFER_SIZE - 1) {
          ERR_RETURN("Can't read comp_label.");
      }
    }

    for (++j; line[j] == ' '; j++);

    for (; line[j] != '\n'; j++) {
      units[m++] = line[j];
      if (m >= BUFFER_SIZE - 1) {
          ERR_RETURN("Can't read unit_label.");
      }
    }

    if (i != *num_comp - 1) {
      data_labels[n++] = '.';
      units[m++] = '.';
      if (m >= BUFFER_SIZE - 1) {
        ERR_RETURN("Can't read comp_label");
      }
    }
  }
  data_labels[n] = '\0';
  units[m] = '\0';

  for (node = 0; node < num_data; node++) {

    num = fscanf (fp, "%d", &id);

    for (offset = i = 0; i < *num_comp; i++) {

      num_vals = comp_list[i];

      for (j = 0; j < num_vals; j++) {

	num = fscanf (fp, "%f",  &value);

	if (num != 1) {
	  ERR_RETURN("Can't read step data value");
	}
	data[node * num_vals + offset + j] = value;
      }

      offset += num_data * num_vals;
    }
  }

  fgets(line, BUFFER_SIZE, fp);

  return 1;
}

/**********************************************************************
 *
 *	function name:  read_node_cell_data
 *
 *      description:    read data of specified step
 *
 *	return:		1: success 
 *			0: failure
 *
 **********************************************************************/
static int
read_node_cell_data(
FILE		*fp,		/* file pointer */
int		num_nodes,	/* number of nodes */
int		num_cells,	/* number of cells */
OMobj_id	out,		/* object id of output ucd data */
OMobj_id	id,		/* object id of parent object */
struct cell	*Top,		/* cell information */
int		time_step	/* time step: if 0, output is single field */
)
{
    char	line[BUFFER_SIZE];
    char	labels[BUFFER_SIZE], units[BUFFER_SIZE];
    char	label[BUFFER_SIZE], unit[BUFFER_SIZE];
    int		i, j, k, m, cs, cnt, type, size;
    int		*comp_list, num_comp, offset, veclen;
    int		num_node_data, num_cell_data, dum, num;
    float	*data, *out_data;
    struct cell	*p;
    OMobj_id	cell_set;
    OMobj_id	node_data_id, cell_data_id;
    int		first_num_comp, first_cell_comp;

    /******************************************/
    /* read number of node data and cell data */
    /******************************************/
    fgets(line, sizeof(line), fp);
    num = sscanf(line, "%d%d", &num_node_data, &num_cell_data);
    if (num != 2) {
	ERR_RETURN("Can't read num_ndata and num_cdata.");
    }

    /******************/
    /* read node data */
    /******************/
    if (num_node_data) { /* if node data exist */
	/* allocation */
	data = (float *)malloc(sizeof(float)*num_nodes*num_node_data);
        if (data == NULL) {
	    ERR_RETURN("Can't alloc memory.(node_data)");
	}
	comp_list = (int *)malloc(sizeof(int)*num_node_data);
        if (comp_list == NULL) {
	    free(data);
	    ERR_RETURN("Can't alloc memory.(comp_list)");
	}

	/* read node data */
	if (read_data(fp, num_nodes, num_node_data, data, labels,
			   units, &num_comp, comp_list) != 1) {
	    free(data);
	    free(comp_list);
	    ERR_RETURN("Can't read node data.");
	}

	/* check or set number of components */
	if (time_step >= 2) {
	    if (FLDget_node_data_ncomp(out, &first_num_comp) != 1) {
		first_num_comp = 0;
	    }
	    if (num_comp != first_num_comp) {
	        free(data);
	        free(comp_list);
	        ERR_RETURN("Can't read this data under Store All Step mode because numbers of node components don't match between steps.")
	    }
	} else {
	    FLDset_node_data_ncomp(out, num_comp);
	}

	/* get time_node_data object id */
	if (time_step >= 1) { 	/* in case of time-dependent field */
	    if (FLDget_time_node_data(out, time_step-1, &node_data_id) != 1) {
	    	ERR_RETURN("Can't get time_node_data id.");
	    }
	} else { 	/* in case of single step field */
	    node_data_id = out;
	}

	for (i = offset = 0; i < num_comp; i++) {

	    /* get a label of node data component */
	    if (!get_label(labels,'.', i, label)) {
		sprintf(label, "node_data_%d", i );
	    }

	    /* get a unit of node data component */
	    if (!get_label(units,'.', i, unit)) {
		sprintf(unit, " ");
	    }

	    /* set node data component information */
	    veclen = comp_list[i];
	    FLDset_node_data_comp(node_data_id, i, veclen, label, unit);

	    /* set node data value */
	    FLDset_node_data(node_data_id, i, (char *)&data[offset],
			DTYPE_FLOAT, veclen*num_nodes, OM_SET_ARRAY_COPY );

	    offset += (veclen * num_nodes);
	}
	free(data);
	free(comp_list);

    } else {
	if (time_step >= 2) {
	    if (FLDget_node_data_ncomp(out, &first_num_comp) == 1) {
	       if (first_num_comp > 0) {
	           ERR_RETURN("Can't read this data under Store All Step mode because numbers of node components don't match between steps.")
               }
	    }
        }
    }


    /******************/
    /* read cell data */
    /******************/
    if (num_cell_data) { /* if cell data exist */
	/* allocation */
	data = (float *)malloc(sizeof(float)*num_cells*num_cell_data);
	if (data == NULL) {
	    ERR_RETURN("Can't alloc memory.(node_data)");
	}
	comp_list = (int *)malloc(sizeof(int)*num_cell_data);
	if (comp_list == NULL) {
	    free(data);
	    ERR_RETURN("Can't alloc memory.(comp_list)");
	}

	/* read cell data */
	if (read_data(fp, num_cells, num_cell_data, data, labels,
		       units, &num_comp, comp_list) != 1) {
	    free(data);
            free(comp_list);
            ERR_RETURN("Can't read cell data.");
        }

	/* set cell data */
	for (p = Top, cs = 0; p != NULL; p = p->next, cs++) {

	    /* get cell_set id */
	    if (FLDget_cell_set(out, cs, &cell_set) != 1) {
		ERR_RETURN("Can't get cell_set.")
	    }

	    /* set number of components */
	    if (time_step >= 2) {
		if (FLDget_cell_data_ncomp(cell_set, &first_cell_comp) != 1) {
		    first_cell_comp = 0;
		}
		if (num_comp != first_cell_comp) {
		    free(data);
		    free(comp_list);
		    ERR_RETURN("Can't read this data under Store All Step mode because numbers of cell components don't match between steps.")
		}
	    } else {
		FLDset_cell_data_ncomp(cell_set, num_comp);
	    }

	    /* get time-dependent cell data id */
	    if (time_step >= 1) {
	        if (FLDget_time_cell_data(cell_set, time_step-1,
			&cell_data_id) != 1) {
		    free(data);
		    free(comp_list);
		    ERR_RETURN("Can't get time_cell_data id.");
		}
	    } else {
	    	cell_data_id = cell_set;
	    }

	    for (i = offset = 0; i < num_comp; i++) {

	        /* get a label of cell data component */
	        if (!get_label(labels,'.', i, label)) {
	            sprintf(label, "cell_data_%d", i);
	        }

	        /* get a unit of cell data component */
	        if (!get_label(units,'.', i, unit)) {
	            sprintf(unit, " ");
	        }

	        /* set node data compornent information */
	        veclen = comp_list[i];
		if (FLDset_cell_data_comp(cell_data_id, i, veclen, label,
			unit ) != 1) {
		    free(data);
		    free(comp_list);
		    ERR_RETURN("Can't set cell_data_comp.");
                }

		/* set cell data value */
		type = DTYPE_FLOAT;
		if (FLDget_cell_data(cell_data_id, i, &type,
			(char **)&(out_data), &size, OM_GET_ARRAY_WR) != 1) {
		    free(data);
		    free(comp_list);
		    ERR_RETURN("Can't get cell_data array.");
                }

		for (j = cnt = 0; j < p->cell_num; j++) {
		    k = p->cell_no[j];
		    for (m = 0; m < veclen; m++, cnt++) {
			out_data[cnt] = data[offset+veclen*k+m];
		    }
		}
		offset += (veclen * num_cells);
		ARRfree(out_data);
	    }
	}

	free(data);
	free(comp_list);

    } else {
	if (time_step >= 2) {
	    if (FLDget_cell_set(out, 0, &cell_set) != 1) {
		ERR_RETURN("Can't get cell_set.")
	    }
	    if (FLDget_cell_data_ncomp(cell_set, &first_cell_comp) == 1) {
		if (first_cell_comp > 0) {
		    ERR_RETURN("Can't read this data under Store All Step mode because numbers of cell components don't match between steps.")
		}
	    }
	}
    }

    return 1;
}

/**********************************************************************
 *
 *	function name:  copy_coord 
 *
 *      description:    copy coord data of the 1st step to other steps. 
 *
 *	return:		1: success 
 *			0: failure
 *
 **********************************************************************/

static int
copy_coord(
OMobj_id	out_all, 	/* ouput time-dependent field */
int		num_steps 	/* number of steps */
) {

    int		i;
    OMobj_id	coord_id;
    float	*coord;
    int		size;

    /* get node data of first step */
    if (FLDget_time_coordinates (out_all, 0, &coord_id) != 1) {
	ERR_RETURN("Can't get coord_id of output time-dependent field.");
    }
    if (FLDget_coord(coord_id, &coord, &size, OM_GET_ARRAY_RD) != 1) {
        ERR_RETURN("Can't get coordinate.");
    }

    /* set coordinate of other steps */
    for (i = 1; i < num_steps; i++) {
        if (FLDget_time_coordinates (out_all, i, &coord_id) != 1) {
	    ERR_RETURN("Can't get coord_id of output time-dependent field.");
        }
        if (FLDset_coord(coord_id, coord, size, OM_SET_ARRAY_STATIC) != 1) {
            ERR_RETURN("Can't get coordinate.");
        }
    }

    ARRfree(coord);

    return(1);
}


/**********************************************************************
 *
 *	function name:  copy_node_cell 
 *
 *      description:    copy node and/or cell data of the 1st step to 
 *			other steps.
 *
 *	return:		1: success 
 *			0: failure
 *
 **********************************************************************/

static int
copy_node_cell(
OMobj_id	out_all, 	/* ouput time-dependent field */
int		num_steps,	/* number of steps */
int		num_nodes,
int		num_cells
) {

    int		node_ncomp;
    OMobj_id	node_data_id, *data_id_list;
    int		i, j;
    char	unit[BUFFER_SIZE], label[BUFFER_SIZE];
    int		veclen, type, data_size;
    float	*node_data, *cell_data;
    int		step;
    int		ncell_sets, cell_ncomp;
    OMobj_id	cell_set_id, cell_data_id;


    /******************/
    /* copy node data */
    /******************/
    node_ncomp = 0;
    if (FLDget_node_data_ncomp (out_all, &node_ncomp) != 1) {
        node_ncomp = 0;
    }

    if (node_ncomp > 0) {
	data_id_list = (OMobj_id *) malloc(num_steps * sizeof(OMobj_id));
	if (!data_id_list) {
	    ERR_RETURN("Can't alloc data_id_list.");
        }
        for (i = 0; i < num_steps; i++) {
	    if (FLDget_time_node_data(out_all, i, &node_data_id) != 1) {
		free(data_id_list);
		ERR_RETURN("Can't get first time_node_data id.");
	    }
	    data_id_list[i] = node_data_id;
        }

	for (i = 0; i < node_ncomp; i++) {
	    if (FLDget_node_data_units(data_id_list[0], i, unit,
			BUFFER_SIZE) != 1) {
		free(data_id_list);
		ERR_RETURN("Can't get first node_data_units.");
            }
	    if (FLDget_node_data_label(data_id_list[0], i, label,
			BUFFER_SIZE) != 1) {
		free(data_id_list);
		ERR_RETURN("Can't get first node_data_label.");
            }
	    if (FLDget_node_data_veclen(data_id_list[0], i, &veclen) != 1) {
		free(data_id_list);
		ERR_RETURN("Can't get first node_data_velcen.");
            }
	    if (FLDget_node_data(data_id_list[0], i, &type,
		    (char **)&node_data, &data_size, OM_GET_ARRAY_RD) != 1) {
		free(data_id_list);
		ERR_RETURN("Can't get first node_data.");
            }

	    for (step = 1; step < num_steps; step++) {
		if (FLDset_node_data_comp(data_id_list[step], i, veclen,
			label, unit) != 1) {
		    free(data_id_list);
		    ERR_RETURN("Can't set node_data_comp.");
                }
		if (FLDset_node_data(data_id_list[step], i, (char *)node_data,
			type, data_size, OM_SET_ARRAY_STATIC) != 1) {
		    free(data_id_list);
		    ERR_RETURN("Can't set node_data.");
                }
            }
        }
        free(data_id_list);
    }

    /******************/
    /* copy cell data */
    /******************/
    if (FLDget_ncell_sets(out_all, &ncell_sets) != 1) {
	ERR_RETURN("Can't get ncell_sets.");
    }
    if (FLDget_cell_set(out_all, 0, &cell_set_id) != 1) {
	ERR_RETURN("Can't get cell_set id.");
    }
    cell_ncomp = 0;
    if (FLDget_cell_data_ncomp(cell_set_id, &cell_ncomp) != 1) {
	return(1);	/* no cell data */
    }
    if (cell_ncomp == 0) {
	return(1);
    }

    for (j = 0; j < ncell_sets; j++) {
	if (FLDget_cell_set(out_all, j, &cell_set_id) != 1) {
	    ERR_RETURN("Can't get cell_set id.");
	}


	for (i = 0; i < cell_ncomp; i++) {
	    if (FLDget_time_cell_data(cell_set_id, 0, &cell_data_id) != 1) {
	        ERR_RETURN("Can't get cell_data id.");
	    }
	    if (FLDget_cell_data_units(cell_data_id, i, unit,
			    BUFFER_SIZE) != 1) {
		ERR_RETURN("Can't get first cell_data_units.");
	    }
	    if (FLDget_cell_data_label(cell_data_id, i, label,
			    BUFFER_SIZE) != 1) {
		ERR_RETURN("Can't get first cell_data_label.");
	    }
	    if (FLDget_cell_data_veclen(cell_data_id, i, &veclen) != 1) {
		ERR_RETURN("Can't get cell_data_veclen.");
	    }
	    if (FLDget_cell_data(cell_data_id, i, &type,
		    (char **)&cell_data, &data_size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("Can't get first node_data.");
	    }

	    for (step = 1; step < num_steps; step++) {
		if (FLDget_time_cell_data(cell_set_id, step,
					    &cell_data_id) != 1) {
		    ERR_RETURN("Can't get cell_data_id.");
		}
		if (FLDset_cell_data_comp(cell_data_id, i, veclen,
			label, unit) != 1) {
		    ERR_RETURN("Can't set cell_data_comp.");
		}
		if (FLDset_cell_data(cell_data_id, i, (char *)cell_data,
			    type, data_size, OM_SET_ARRAY_STATIC) != 1) {
		    ERR_RETURN("Can't set cell_data.");
		}
	    }
	}
    }
    return(1);
}


/**********************************************************************
 *
 *	function name:  isdigit 
 *
 *      description:    check if input string is numerical value or not 
 *
 *	return:		1: numerical value 
 *			0: not numerical value 
 *
 **********************************************************************/

static int
isvalue(char *str_array)
{

   int i,len,flag;
   flag = 1;
   len = strlen(str_array);
   for(i = 0; i < len; i++)
   {
     if(( str_array[i] < 0x30 &&  str_array[i] != 0x2e)  ||
            ( str_array[i] > 0x39 ))
     {
       flag = 0;
       break;
     }
   }
   return(flag);
}


/**********************************************************************
 *
 *	function name:  set_time 
 *
 *      description:    set time in time-dependent field. 
 *
 *	return:		1: success 
 *			0: failure
 *
 **********************************************************************/

static int
set_time(
OMobj_id	out_all,
int	num_steps,
char	(*all_comments)[BUFFER_SIZE]	/* comments of all steps */
)
{

    int		i;
    double	*time;
    char	comment[BUFFER_SIZE];
    int		float_flag;
    OMobj_id 	set_id;
    int		ncell_sets;


    time = (double *) malloc(num_steps * sizeof(double));
    if (time == NULL) {
	ERR_RETURN("Can't alloc time.");
    }

    float_flag = 1;
    for (i = 0; i < num_steps; i++) {
	strcpy(comment, all_comments[i]);
	if (isvalue(comment) != 1) {
            float_flag = 0;
	    break;
	}
	if (sscanf(comment, "%lf", &time[i]) != 1) {
            float_flag = 0;
	    break;
        }
    }

    if (float_flag == 0) {
		/* if any comment is not numerical value, sequential
                   numbers are set into time */
	for (i = 0; i < num_steps; i++) {
	    time[i] = i;
	}
    }

    if (FLDset_time_values(out_all, time, num_steps,
				OM_SET_ARRAY_COPY) != 1) {
        free(time);
	ERR_RETURN("Can't set time.");
    }

    if (FLDget_ncell_sets(out_all, &ncell_sets) != 1) {
	free(time);
	ERR_RETURN("Can't get ncell_sets");
    }

    for (i = 0; i < ncell_sets; i++) {
	if (FLDget_cell_set(out_all, i, &set_id) != 1) {
	    free(time);
	    ERR_RETURN("Can't get cell_set id.");
	}
	if (FLDset_time_values(set_id, time, num_steps,
				    OM_SET_ARRAY_COPY) != 1) {
	    free(time);
	    ERR_RETURN("Can't set time.");
        }
    }
    free(time);

    return(1);
}


/**********************************************************************
 *
 *	function name:  read_all_step
 *
 *      description:    read and set data of all steps
 *
 *	return:		1: success 
 *			0: failure
 *
 **********************************************************************/

static int
read_all_step(
FILE 	*fp,
int	num_steps,
char	*cycle_type,
OMobj_id	out_all, 	/* ouput time-dependent field */
char	(*all_comments)[BUFFER_SIZE],
struct  cell	**Top,		/* cell information */
OMobj_id	id		/* object id of parent object */
) {

char	comment[BUFFER_SIZE];
int	num_nodes, num_cells;
int	i;
int	step;
char	err_mes[128];


    /********************************/
    /* read and set first step data */
    /********************************/
    /* set num_steps */
    if (FLDset_time_nsteps(out_all, num_steps) != 1) {
	ERR_RETURN("Can't set num_steps.");
    }

    /* read and set first step */
    if (read_step_comment(fp, 1, comment) != 1) {
	ERR_RETURN("Can't find 'step' or step number is incorrect.");
    }
    strcpy(all_comments[0], comment);

    if (read_geom_data(fp, &num_nodes, &num_cells, out_all, Top, 1,
      		num_steps) != 1) {
	delete_cell_inf(Top);
	ERR_RETURN("Can't set geometry data of the first step.");
    }

    if (read_node_cell_data(fp, num_nodes, num_cells,
				    out_all, id, *Top, 1) != 1) {
	delete_cell_inf(Top);
	ERR_RETURN("Can't set node or cell data of the first step.");
    }

    /**************************************************/
    /* read and set second and after second step data */
    /**************************************************/

    for (step = 2; step <= num_steps; step++) {
	if (read_step_comment(fp, step, comment) != 1) {
            sprintf(err_mes,
		"Error in step %d: Can't find 'step' or step number is incorrect.",
	     	step);
	    ERR_RETURN(err_mes);
	}
	strcpy(all_comments[step-1], comment);

	if (strcmp(cycle_type, "data_geom") == 0 ||
		strcmp(cycle_type, "geom") == 0) {
    	    if (read_geom_data(fp, &num_nodes, &num_cells, out_all, Top,
			step, num_steps) != 1) {
	        delete_cell_inf(Top);
                sprintf(err_mes,
			"Error in step %d: Can't set geometry data.",
			step);
	        ERR_RETURN(err_mes);
	    }
        }
	if (strcmp(cycle_type, "data_geom") == 0 ||
		strcmp(cycle_type, "data") == 0) {
	    if (read_node_cell_data(fp, num_nodes, num_cells,
				    out_all, id, *Top, step) != 1) {
		delete_cell_inf(Top);
                sprintf(err_mes,
			"Error in step %d: Can't set node or cell data",
			step);
	        ERR_RETURN(err_mes);
	    }
        }
    }

    if (strcmp(cycle_type, "data") == 0) {
	if (copy_coord(out_all, num_steps) != 1) {
	    ERR_RETURN("Can't copy coordinates.");
        }
    }

    if (strcmp(cycle_type, "geom") == 0) {
	if (copy_node_cell(out_all, num_steps, num_nodes, num_cells) != 1) {
	    ERR_RETURN("Can't copy node and cell data.");
	}
    }

    if (set_time(out_all, num_steps, all_comments) != 1) {
	ERR_RETURN("Can't set time.");
    }

    return(1);
}


/**********************************************************************
 *
 *	function name: 	extract_one_step 
 *
 *      description:    extract one step data from time-dependent field
 *			and set it into single field
 *
 *	return:		1: success 
 *			0: failure
 *
 **********************************************************************/

static int
extract_one_step(
OMobj_id	out_all,	/* time-dependent field data */
int		step,		/* step number (count from 0) */
OMobj_id	out		/* output single field */
)
{

    int		nnodes, node_ncomp, size, type, ncell_sets, ncells;
    float	*coord, *node_data, *cell_data;
    int		i, j;
    char	name[10], label[BUFFER_SIZE], units[BUFFER_SIZE];
    OMobj_id	all_coord_id, all_node_data_id, cell_set_id,
		all_cell_set_id, all_cell_data_id;
    int		*node_conn;
    int		cell_ncomp, veclen;
    float	*props, *all_props;


    if (FLDget_nnodes(out_all, &nnodes) != 1) {
	ERR_RETURN("Can't get nnodes.");
    }
    if (FLDset_nnodes(out, nnodes) != 1) {
	ERR_RETURN("Can't set nnodes.");
    }

    if (FLDset_nspace(out, 3) != 1) {
	ERR_RETURN("Can't set nspace.");
    }

    if (FLDget_time_coordinates(out_all, step, &all_coord_id) != 1) {
	ERR_RETURN("Can't get coord_id.");
    }
    if (FLDget_coord(all_coord_id, &coord, &size, OM_GET_ARRAY_RD) != 1) {
	ERR_RETURN("Can't get coord.");
    }
    if (FLDset_coord(out, coord, size, OM_SET_ARRAY_STATIC) != 1) {
	ERR_RETURN("Can't set coord.");
    }
    ARRfree(coord);

    if (FLDget_node_data_ncomp(out_all, &node_ncomp) != 1) {
	node_ncomp = 0;
    }

    if (node_ncomp > 0) {
	if (FLDset_node_data_ncomp(out, node_ncomp) != 1) {
	    ERR_RETURN("Can't get node_data_ncomp.");
	}

	if (FLDget_time_node_data(out_all, step, &all_node_data_id) != 1) {
	    ERR_RETURN("Can't get node_data_id.");
	}
	if (FLDset_node_data_ncomp(out, node_ncomp) != 1) {
	    ERR_RETURN("Can't set node_data_ncomp.");
	}

	if (FLDget_time_node_data(out_all, step, &all_node_data_id) != 1) {
	    ERR_RETURN("Can't get node_data_id.");
	}

	for (i = 0; i < node_ncomp; i++) {
	    if (FLDget_node_data_veclen(all_node_data_id, i, &veclen) != 1) {
		ERR_RETURN("Can't get node_data_veclen.");
	    }
	    if (FLDget_node_data_units(all_node_data_id, i, units,
				    BUFFER_SIZE) != 1) {
		ERR_RETURN("Can't get node_data_units.");
	    }
	    if (FLDget_node_data_label(all_node_data_id, i, label,
				    BUFFER_SIZE) != 1) {
		ERR_RETURN("Can't get node_data_label.");
	    }

	    if (FLDset_node_data_comp(out, i, veclen, label, units) != 1) {
		ERR_RETURN("Can't set node_data_comp.");
	    }

	    if (FLDget_node_data(all_node_data_id, i, &type, (char **)&node_data,
			     &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("Can't get node_data.");
	    }
	    if (FLDset_node_data(out, i, (char *)node_data, type, size,
			    OM_SET_ARRAY_STATIC) != 1) {
		ERR_RETURN("Can't set node_data.");
	    }
	    ARRfree(node_data);
	}
    }

    if (FLDget_ncell_sets(out_all, &ncell_sets) != 1) {
	ERR_RETURN("Can't get ncell_sets.");
    }
    if (FLDset_ncell_sets(out, ncell_sets) != 1) {
	ERR_RETURN("Can't set ncell_sets.");
    }

    for (i = 0; i < ncell_sets; i++) {
	if (FLDget_cell_set(out_all, i, &all_cell_set_id) != 1) {
	    ERR_RETURN("Can't get cell_set id.");
	}
	if (FLDget_cell_set(out, i, &cell_set_id) != 1) {
	    ERR_RETURN("Can't get cell_set id.");
	}

	if (FLDget_cell_set_name(all_cell_set_id, name, 10) != 1) {
	    ERR_RETURN("Can't get cell_set id");
	}
	if (FLDset_cell_set(cell_set_id, name) != 1) {
	    ERR_RETURN("Can't set cell_set.");
	}

	if (FLDget_ncells(all_cell_set_id, &ncells) != 1) {
	    ERR_RETURN("Can't get ncells.");
        }
	if (FLDset_ncells(cell_set_id, ncells) != 1) {
	    ERR_RETURN("Can't set ncells.");
	}

        if (FLDget_node_connect(all_cell_set_id, &node_conn, &size,
		OM_GET_ARRAY_RD) != 1) {
	    ERR_RETURN("Can't get node_connect.");
	}
        if (FLDset_node_connect(cell_set_id, node_conn, size,
			    OM_SET_ARRAY_STATIC) != 1) {
	    ERR_RETURN("Can't get node_connect.");
	}
	ARRfree(node_conn);

	if (FLDset_cell_nprops (cell_set_id, 1) != 1) {
	    ERR_RETURN("Can't set nprops.");
	}
	if (FLDget_cell_props(all_cell_set_id, (float **)&all_props, &size,
		    OM_GET_ARRAY_RD) != 1) {
	    ERR_RETURN("Can't get cell props of time-dependent field.");
	}
	if (FLDget_cell_props(cell_set_id, (float **)&props, &size,
		    OM_GET_ARRAY_WR) != 1) {
	    ERR_RETURN("Can't get cell props.");
	}
	props[0] = all_props[0];
	ARRfree(props);
	ARRfree(all_props);

	if (FLDget_time_cell_data(all_cell_set_id, step,
			&all_cell_data_id) != 1) {
	    ERR_RETURN("Can't get time_cell_data id.");
	}

	cell_ncomp = 0;
	if (FLDget_cell_data_ncomp(all_cell_data_id, &cell_ncomp) != 1) {
	    return(1);
	}
	if (FLDset_cell_data_ncomp(cell_set_id, cell_ncomp) != 1) {
	    ERR_RETURN("Can't set cell_data_ncomp.");
	}

        for (j = 0; j < cell_ncomp; j++) {
	    if (FLDget_cell_data_veclen(all_cell_data_id, j, &veclen) != 1) {
		ERR_RETURN("Can't get cell_data_veclen.");
	    }
	    if (FLDget_cell_data_units(all_cell_data_id, j, units,
				    BUFFER_SIZE) != 1) {
		ERR_RETURN("Can't get node_cell_units.");
	    }
	    if (FLDget_cell_data_label(all_cell_data_id, j, label,
				    BUFFER_SIZE) != 1) {
		ERR_RETURN("Can't get cell_data_label.");
	    }
	    if (FLDset_cell_data_comp(cell_set_id, j, veclen, label,
				units) != 1) {
		ERR_RETURN("Can't set cell_data_comp.");
	    }

	    if (FLDget_cell_data(all_cell_data_id, j, &type,
			(char **)&cell_data, &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("Can't get cell_data");
	    }
	    if (FLDset_cell_data(cell_set_id, j, (char *)cell_data, type, size,
			    OM_SET_ARRAY_STATIC) != 1) {
		ERR_RETURN("Can't set cell_data");
	    }
	}
    }
    return(1);
}


     
/**********************************************************************
 *
 *	function name:  DVread_ucd2_update
 *
 *      description:	update function (method)
 *
 *	return:		1: 
 *			0: 
 *
 **********************************************************************/
int
DVread_ucd2_update(
OMobj_id	id,
OMevent_mask	event_mask,
int		seq_num
)
{
    char	cycle_type[BUFFER_SIZE];
    char	*cycle_ptr;
    int		new_format;
    char   	comment[BUFFER_SIZE];
    int		num_steps, num_nodes, num_cells;
    char  	*filename = NULL, err_mes[128];
    int  	current_step, chg_step, one_time;
    OMobj_id	out, chk_id;
    char	line[BUFFER_SIZE];
    FILE	*FilePtr;
    long	*FilePos;
    struct cell	*Top;
    int         ncell_sets, ncells;
    OMobj_id    cell_set;
    int         i;
    int		store_all, store_all_visible;
    char	(*all_comments)[BUFFER_SIZE];
    OMobj_id	out_all;
    int		file_changed, store_changed;


    /*************************************/
    /* check if filename is input or not */
    /*************************************/

    chk_id = OMfind_subobj(id, OMstr_to_name("filename"), OM_OBJ_RD);
    if (OMis_null_obj(chk_id)) {
        ERR_RETURN("Can't get 'filename' object id.");
    }
    if (OMchanged(chk_id, seq_num)) {
        file_changed = 1;
    } else {
        file_changed = 0;
    }

    /*******************************************************/
    /* check if store_all is changed or not  and its value */
    /*******************************************************/

    chk_id = OMfind_subobj(id, OMstr_to_name("store_all"), OM_OBJ_RD);
    if (OMis_null_obj(chk_id)) {
        ERR_RETURN("Can't get 'store_all' object id.");
    }
    if (OMchanged(chk_id, seq_num)) {
        store_changed = 1;
    } else {
        store_changed = 0;
    }

    /* get store_all */
    if (OMget_name_int_val(id, OMstr_to_name("store_all"),
	     &store_all) != 1) {
        ERR_RETURN("Can't get Store All Steps.");
    }

    /***************************************************/
    /* process in case filname or store_all is changed */
    /***************************************************/

    if (file_changed) {
	/* set store_all = 0 */
	OMset_name_int_val(id, OMstr_to_name("store_all"), 0);
        store_all = 0;
	OMset_name_int_val(id, OMstr_to_name("store_all_visible"), 0);
	store_all_visible = 0;
    }

    if (store_changed && store_all == 0) {
	/* set store_all_visible = 0 */
	OMset_name_int_val(id, OMstr_to_name("store_all_visible"), 0);
        store_all_visible = 0;
    }

    if (file_changed || store_changed) {
	/* reset current_step and total */
	OMset_name_int_val(id, OMstr_to_name("current_step"), 0);
	OMset_name_int_val(id, OMstr_to_name("total"), 0 );

	/* free cell information */
	if (OMget_name_ptr_val(id, OMstr_to_name("top"), (void *)
		&Top, 0) != 1) {
	    ERR_RETURN("Can't get saved top.");
	}
	delete_cell_inf(&Top);
	if (OMset_name_ptr_val(id, OMstr_to_name("top"), Top, 0) != 1) {
	    ERR_RETURN("Can't save top.");
	}

	/* get object ID of UCD ouput object */
	out = OMfind_subobj(id, OMstr_to_name("out"), OM_OBJ_RW);
	if (OMis_null_obj(out)) {
	    ERR_RETURN("Can't get 'out' object id.");
	}
	out_all = OMfind_subobj(id, OMstr_to_name("out_all"), OM_OBJ_RW);
	if (OMis_null_obj(out_all)) {
	    ERR_RETURN("Can't get 'out_all' object id.");
	}

        /* get input filename */
        filename = NULL;
        if (OMget_name_str_val(id, OMstr_to_name("filename"),
			&filename, 0) != 1) {
	    ERR_RETURN("Can't get filename.");
	}
        if (filename == NULL) {
	    ERR_RETURN("Please select filename.");
        }

	/* initialize output UCD data */
	reset_out_data(out, 0);
	reset_out_data(out_all, 1);

	/* open file */
	if (OMget_name_ptr_val(id, OMstr_to_name("file_ptr"),
		(void *) &FilePtr, 0) != 1) {
	    ERR_RETURN("Can't get file_ptr");
        }
	if (FilePtr != NULL) {
	    fclose(FilePtr);
	    FilePtr = NULL;
	    if (OMset_name_ptr_val(id, OMstr_to_name("file_ptr"),
			FilePtr, 0) != 1) {
	        free(filename);
		ERR_RETURN("Can't save file_ptr.");
	    }
	}
	FilePtr = FILEfopen(filename, SIO_R_TXT);
	if (FilePtr == NULL) {
	    sprintf(err_mes, "Can't open file[%s]", filename);
	    free(filename);
	    ERR_RETURN(err_mes);
	}

	/* read num_steps and cycle_type */
	cycle_type[0] = '\0';
	if (read_header(filename, FilePtr, &num_steps, cycle_type) == 0) {
	    fclose(FilePtr);
	    free(filename);
	    ERR_RETURN("Can't read header section.");
	}
	/* store cycle_type */
	if (OMset_name_str_val(id, OMstr_to_name("cycle_type"),
		cycle_type) != 1) {
	    ERR_RETURN("Can't save cycle_type.");
	}

	/* in case of new (time-dependent) format */
	if (strcmp(cycle_type, "data") == 0 ||
	    strcmp(cycle_type, "geom") == 0 ||
	    strcmp(cycle_type, "data_geom") == 0) {

	    new_format = 1;

            if (file_changed) {
		/* set store_all_visible = 1 */
		OMset_name_int_val(id, OMstr_to_name("store_all_visible"), 1);
		store_all_visible = 1;
            }

	    if (num_steps < 1) {
		fclose(FilePtr);
		free(filename);
		ERR_RETURN("num step must be greater than 0");
	    }

	    /* in case data of all steps will be stored */
	    if (store_all == 1) {

		/* free and malloc all_comments */
		if (OMget_name_ptr_val(id, OMstr_to_name("all_comments"),
				(void *) &all_comments, 0) != 1) {
		    ERR_RETURN("Can't restore all_comments.");
                }

	  	if (all_comments != NULL) {
		    free (all_comments);
		}
		all_comments = (char (*)[BUFFER_SIZE])
					malloc(num_steps * BUFFER_SIZE);
		if (OMset_name_ptr_val(id, OMstr_to_name("all_comments"),
				all_comments, 0) != 1) {
		    ERR_RETURN("Can't store all_comments.");
		}

		if (all_comments == NULL) {
		    ERR_RETURN("Can't allocate all_comments");
		}

	        /* read and set data of all steps */
		if (read_all_step(FilePtr, num_steps, cycle_type, out_all,
				all_comments, &Top, id) != 1 ) {
	            fclose (FilePtr);
		    free(filename);
		    return(0);
                }

		/* store Top (cell information) */
		if (OMset_name_ptr_val(id, OMstr_to_name("top"), Top, 0) != 1) {
		    ERR_RETURN("Can't save cell_info");
		}


		if (extract_one_step(out_all, 0, out) != 1) {
	            fclose (FilePtr);
		    free(filename);
		    ERR_RETURN("Can't extract one step.");
	        }

		/* set comment to object */
		if (OMset_name_str_val(id, OMstr_to_name("step_title"),
			all_comments[0]) != 1) {
		    ERR_RETURN("Can't set step_title.");
                }

	    /* in case data of single step will be stored */
	    } else {

		/* check step and get step comment string */
		comment[0] = '\0';
		if (read_step_comment(FilePtr, 1, comment) != 1) {
		    fclose(FilePtr);
		    free(filename);
		    ERR_RETURN("Can't find 'step' or step number is incorrect.");
		}

		/* set comment to object */
		if (OMset_name_str_val(id, OMstr_to_name("step_title"),
			comment) != 1) {
		    ERR_RETURN("Can't set step_title.");
                }

		/* read and set geometry data of the first step */
		if (read_geom_data(FilePtr, &num_nodes, &num_cells, out,
			    &Top, 0, 0) != 1) {
		    delete_cell_inf(&Top);
		    OMset_name_ptr_val(id, OMstr_to_name("top"), Top, 0);
		    fclose(FilePtr);
		    free(filename);
		    ERR_RETURN("Can't set geometry data of the first step.");
		}
		/* store Top (cell information) */
		if (OMset_name_ptr_val(id, OMstr_to_name("top"), Top, 0) != 1) {
		    ERR_RETURN("Can't save cell_info");
		}


		/* read and set node and cell data of the first step */
		if (read_node_cell_data(FilePtr, num_nodes, num_cells,
				    out, id, Top, 0) != 1 ) {
		    fclose(FilePtr);
		    free(filename);
		    ERR_RETURN("Can't set node or cell data of the first step.");
		}

		/* get file pointer of beginning of each step */
		if (OMget_name_ptr_val(id, OMstr_to_name("file_pos"),
			(void *) &FilePos, 0) != 1) {
		    ERR_RETURN("Can't restore file_pos.");
	 	}
		if (FilePos != NULL) {
		    free(FilePos);
		    FilePos = NULL;
		    if (OMset_name_ptr_val(id, OMstr_to_name("file_pos"),
			    FilePos, 0) != 1) {
			ERR_RETURN("Can't store file_pos.");
		    }
		}
		if (get_step_file_pos(FilePtr, num_steps, cycle_type,
			&FilePos) != 1) {
		    delete_cell_inf(&Top);
		    fclose(FilePtr);
		    free(filename);
		    ERR_RETURN("Can't get position of each step.");
		}

		/* store file position */
		if (OMset_name_ptr_val(id, OMstr_to_name("file_pos"),
				FilePos, 0) != 1) {
		    ERR_RETURN("Can't store file_pos.");
		}
            }

	/* in case of old format */
	} else {
	    new_format = 0;
	    num_steps = 1;

 	    fclose(FilePtr);
	    FilePtr = NULL;
	    FUNCread_ucd(filename, out);
	    if (OMset_name_str_val(id, OMstr_to_name("step_title"), "")
			!= 1) {
		ERR_RETURN("Can't set step_title.");
	    }
	}

	/* set total number of steps to the parameter object */
	if (OMset_name_int_val(id, OMstr_to_name("total"), num_steps)
		    != 1) {
	    ERR_RETURN("Can't set total.");
	}

	/* set 1 to current step when new file is read */
	current_step = 1;
	if (OMset_name_int_val(id, OMstr_to_name("current_step"),
		current_step) != 1) {
	    ERR_RETURN("Can't set current_step.");
	}

	/* store file pointer */
	if (OMset_name_ptr_val(id, OMstr_to_name("file_ptr"), FilePtr, 0)
			!= 1) {
	    ERR_RETURN("Can't store file_ptr.");
	}

	/* store new_format */
	if (OMset_name_int_val(id, OMstr_to_name("new_format"), new_format)
			!= 1) {
	    ERR_RETURN("Can't store new_format.");
	}

	free(filename);

	return(1);
    }

    /***************************************/
    /* check paramter and get current step */
    /***************************************/

    /* get store_all */
    if (OMget_name_int_val(id, OMstr_to_name("store_all"),
	     &store_all) != 1) {
        ERR_RETURN("Can't get Store All Steps.");
    }

    /* initilize change_of_step flag */
    chg_step = 0;

    /* check file is open (selected) */
    FilePtr = NULL;
    if (OMget_name_ptr_val(id, OMstr_to_name("file_ptr"),
		(void *) &FilePtr, 0) != 1) {
	ERR_RETURN("Can't restore file_ptr.");
    }
    if (FilePtr == NULL) {
	return (0);
    }

    /* get current_step */
    if (OMget_name_int_val(id, OMstr_to_name("current_step"),
		     &current_step) != 1) {
	ERR_RETURN("Can't get Current Step No.");
    }

    /* get num_steps (total steps) */
    if (OMget_name_int_val(id, OMstr_to_name("total"), &num_steps) != 1) {
	ERR_RETURN("Can't get Total Steps.");
    }

    /* check if forward button was clicked */
    chk_id = OMfind_subobj(id, OMstr_to_name("forward"), OM_OBJ_RD);
    if (OMis_null_obj(chk_id)) {
	ERR_RETURN("Can't get 'forward' object id.");
    }
    if (OMchanged(chk_id, seq_num)) { /* in case forward was clicked */
	/* set current_step (increment) */
	current_step++;
        /* in case current_step is over the total number of steps */
        if (current_step > num_steps) {
            /* return to the first step */
            current_step = 1;
        }

	/* set the value to current_step object */
	if (OMset_name_int_val(id, OMstr_to_name("current_step"),
		current_step) != 1) {
	    ERR_RETURN("Can't set current_step.");
	}

        /* indicate step is changed */
        chg_step = 1;
    }

    /* check if backward button was clicked */
    chk_id = OMfind_subobj(id, OMstr_to_name("backward"), OM_OBJ_RD);
    if (OMis_null_obj(chk_id)) {
	ERR_RETURN("Can't get 'backward' object id.");
    }
    if (OMchanged(chk_id, seq_num)) { /* in case backward was clicked */
	/* set current_step (decrement) */
	current_step--;
        /* in case current_step is less than 1 */
        if (current_step < 1) {
            /* return to the last step */
            current_step = num_steps;
        }

	/* set the value to current_step object */
	if (OMset_name_int_val(id, OMstr_to_name("current_step"),
		current_step) != 1) {
	    ERR_RETURN("Can't set current_step.");
	}

        /* indicate step is changed */
        chg_step = 1;
    }


    /* check if current_step parameter is changed */
    if (chg_step == 0) {
	chk_id = OMfind_subobj(id, OMstr_to_name("current_step"), OM_OBJ_RW);
	if (OMis_null_obj(chk_id)) {
	    ERR_RETURN("Can't get 'current_step' object id.");
	}
        /* in case "current step" parameter is changed */
	if (OMchanged(chk_id, seq_num)) {

	    /* check the validity of current_step */
	    if (current_step < 1 || current_step > num_steps) {
		ERRerror("Read_UCD",1,ERR_ORIG,
			"Step number %d does not exist.", current_step);

		if (current_step < 1) {
		    current_step = 1;
     		}
		if (current_step > num_steps) {
		    current_step = num_steps;
     		}
		if (OMset_name_int_val(id, OMstr_to_name("current_step"),
				current_step) != 1) {
		    ERR_RETURN("Can't set current_step.");
		}
	    }

	    /* indicate current step is changed */
	    chg_step = 1;
	}
    }

    /* get new_format */
    if (OMget_name_int_val(id, OMstr_to_name("new_format"), &new_format)
		!= 1) {
	ERR_RETURN("Can't restore new_format.");
    }

    /* if the format is old and filename is not changed, no data is */
    /* created.                                                     */
    if (new_format == 0) {
	return(1);
    }

    /*************************************/
    /* read and set data of current step */
    /*************************************/

    /* in case current step is changed */
    if (chg_step == 1) {
	/* in case of Store All Step */
        if (store_all == 1) {
	    /* get ID of ucd output object */
	    out = OMfind_subobj(id, OMstr_to_name("out"), OM_OBJ_RW);
	    if (OMis_null_obj(out)) {
		ERR_RETURN("Can't get 'out' object id.");
	    }
	    out_all = OMfind_subobj(id, OMstr_to_name("out_all"), OM_OBJ_RW);
	    if (OMis_null_obj(out_all)) {
		ERR_RETURN("Can't get 'out_all' object id.");
	    }

	    /* get all_comments */
 	    if (OMget_name_ptr_val(id, OMstr_to_name("all_comments"),
				(void *) &all_comments, 0) != 1) {
		ERR_RETURN("Can't restore all_comments.");
	    }

	    /* extract one step data from time-dependent field and set */
            /* it into single field */
	    if (extract_one_step(out_all, current_step-1, out) != 1) {
		ERR_RETURN("Can't extract one step.");
	    }
	    /* set comment to object */
	    if (OMset_name_str_val(id, OMstr_to_name("step_title"),
		    all_comments[current_step-1]) != 1) {
		ERR_RETURN("Can't set step_title.");
	    }

	/* in case of single step */
	} else {

	    /* get Top (cell_info) */
	    if (OMget_name_ptr_val(id, OMstr_to_name("top"), (void *) &Top,
			0) != 1) {
		ERR_RETURN("Can't restore cell_info");
	    }

	    /* get cycle_type */
	    if (OMget_name_str_val(id, OMstr_to_name("cycle_type"),
				(void *) &cycle_ptr, 0) != 1) {
		ERR_RETURN("Can't restore cycle_type");
	    }

	    if (cycle_ptr) {
	        strcpy (cycle_type, cycle_ptr);
	        free (cycle_ptr);
	    }

	    /* get FilePos */
	    if (OMget_name_ptr_val(id, OMstr_to_name("file_pos"),
			(void *) &FilePos, 0) != 1) {
		ERR_RETURN("Can't restore file_pos");
	    }

	    /* set file pointer to specified step */
	    fseek(FilePtr, FilePos[current_step-1], SEEK_SET);

	    /* get ID of ucd output object */
	    out = OMfind_subobj(id, OMstr_to_name("out"), OM_OBJ_RW);
	    if (OMis_null_obj(out)) {
		ERR_RETURN("Can't get 'out' object id.");
	    }

	    /* check step and get step comment string */
	    comment[0] = '\0';
	    if (read_step_comment(FilePtr, current_step, comment) != 1) {
		sprintf(err_mes,
		    "Error in step %d: Can't find 'step' or step number is incorrect.",
		    current_step);
		ERR_RETURN(err_mes);
	    }
	    /* set comment to object */
	    if (OMset_name_str_val(id, OMstr_to_name("step_title"),
			comment) != 1) {
		ERR_RETURN("Can't set step_title.");
	    }

	    /* read geometry data, node data and cell data */
	    if (strcmp(cycle_type, "data") == 0) {
		if (current_step == 1) {
		    if (skip_geom_data(FilePtr, &num_nodes, &num_cells) != 1) {
			ERR_RETURN("Can't skip geometry of the first step.");
		    }
		}
		FLDget_nnodes (out, &num_nodes);

		FLDget_ncell_sets (out, &ncell_sets);
		num_cells = 0;
		for (i = 0; i < ncell_sets; i++) {
		    FLDget_cell_set (out, i, &cell_set);
		    FLDget_ncells (cell_set, &ncells);
		    num_cells += ncells;
		}

		if (read_node_cell_data(FilePtr, num_nodes, num_cells,
			    out, id, Top, 0) != 1) {
		    sprintf(err_mes,
			    "Error in step %d: Can't set node or cell data.",
			    current_step);
		    ERR_RETURN(err_mes);
		}
	    }

	    if (strcmp(cycle_type, "geom") == 0) {
		if (read_geom_data(FilePtr, &num_nodes, &num_cells, out,
			    &Top, 0, 0) != 1) {
		    delete_cell_inf(&Top);
		    OMset_name_ptr_val(id, OMstr_to_name("top"), Top, 0);
		    OMset_name_int_val(id, OMstr_to_name("total"), 0);
		    sprintf(err_mes,
			    "Error in step %d: Can't set geometry data.",
			    current_step);
		    ERR_RETURN(err_mes);
		}
		if (OMset_name_ptr_val(id, OMstr_to_name("top"), Top, 0)
				!= 1) {
		    ERR_RETURN("Can't store top.");
		}

		/* read node and cell data of the first step */
		fseek(FilePtr, FilePos[num_steps], SEEK_SET);
				/* FilePos[num_steps] is the beginning */
				/* of node and cell data of the first step */
		if (read_node_cell_data(FilePtr, num_nodes, num_cells,
			    out, id, Top, 0) != 1) {
		    sprintf(err_mes,
			    "Error in step %d: Can't set node or cell data.",
			    current_step);
		    ERR_RETURN(err_mes);
		}

	    }

	    if (strcmp(cycle_type, "data_geom") == 0) {
		if (read_geom_data(FilePtr, &num_nodes, &num_cells, out,
			    &Top, 0, 0) != 1) {
		    delete_cell_inf(&Top);
		    OMset_name_ptr_val(id, OMstr_to_name("top"), Top, 0);
		    OMset_name_int_val(id, OMstr_to_name("total"), 0);
		    sprintf(err_mes,
			    "Error in step %d: Can't read geometry data.",
			    current_step);
		    ERR_RETURN(err_mes);
		}
		OMset_name_ptr_val(id, OMstr_to_name("top"), Top, 0);

		if (read_node_cell_data(FilePtr, num_nodes, num_cells,
			    out, id, Top, 0) != 1) {
		    sprintf(err_mes,
			    "Error in step %d: Can't set node or cell data.",
			    current_step);
		    ERR_RETURN(err_mes);
		}
	    }
        }
    }
    return 1;
}

/**********************************************************************
 *
 *	function name:  DVread_ucd2_delete
 *
 *      description:	delete function (method)
 *
 *	return:		1: 
 *
 **********************************************************************/
int
DVread_ucd2_delete(
OMobj_id id
)
{
  FILE        *fp;
  long	      *FilePos;
  struct cell *Top;
  char	(*all_comments)[BUFFER_SIZE];

  /* if we have a file open close it */
  if (OMget_name_ptr_val(id, OMstr_to_name("file_ptr"), (void *) &fp, 0)) {
      if (fp != NULL) {
	  fclose(fp);  fp = NULL;
	  OMset_name_ptr_val(id, OMstr_to_name("file_ptr"), fp, 0);
      }
  }

  if (OMget_name_ptr_val(id, OMstr_to_name("top"), (void *) &Top, 0)) {
      if (Top) {
	  delete_cell_inf(&Top);
	  free(Top);  Top = NULL;
	  OMset_name_ptr_val(id, OMstr_to_name("top"), Top, 0);
      }
  }

  if (OMget_name_ptr_val(id, OMstr_to_name("file_pos"), (void *)
			&FilePos, 0)) {
      if (FilePos != NULL) {
  	  free(FilePos);  FilePos = NULL;
	  OMset_name_ptr_val(id, OMstr_to_name("file_pos"), FilePos, 0);
      }
  }

  if (OMget_name_ptr_val(id, OMstr_to_name("all_comments"),
			(void *) &all_comments, 0)) {
      if (all_comments != NULL) {
	  free(all_comments);  all_comments = NULL;
	  OMset_name_ptr_val(id, OMstr_to_name("all_comments"),
		all_comments, 0);
      }
  }
  return(1);
}
